"""
  Hey,don't import me! I am for internal use.

  API功能的具体实现
  -------------------------------------
  Project Tongji-EasyAPI
  models.py
  Copyright (c) 2022 Cinea Zhan. All rights reserved
  www.cinea.com.cn
"""
import json

try:
    from . import models, networkTools
except ImportError or ModuleNotFoundError:
    import models, networkTools
import requests
import time
from typing import Optional

"""
PART.0  解析参数、相关内部工具

"""


def _processArgs(sessionId, cookie) -> tuple:
    if cookie:
        if isinstance(cookie, str):
            cookie = networkTools.parseStrCookie(cookie)
        try:
            sessionId = cookie["sessionid"]
        except:
            raise ValueError("传入的cookie有误")
    elif sessionId:
        cookie = {"sessionid": sessionId, "language": "cn"}
    else:
        raise ValueError("请至少提供一项参数")
    return sessionId, cookie


"""
PART.1  登录、会话、状态类

注释：login仅在Session类中提供
"""


def sessionIdToUserData(sessionId=None, cookie=None, **kw) -> Optional[models.Student]:
    """
    通过已知的sessionId或cookie反查学生信息
    @params: 在sessionId与cookie中任选一项传入即可。
    @return: 返回Student对象。若sessionId无效，则返回None
    """

    if "session" not in kw:
        cookie = _processArgs(sessionId, cookie)[1]

        res = requests.get(f"https://1.tongji.edu.cn/api/sessionservice/session/getSessionUser?_t={networkTools.ts()}",
                           headers=networkTools.headers(), cookies=cookie)
    else:
        session: requests.Session = kw['session']
        res = session.get(f"https://1.tongji.edu.cn/api/sessionservice/session/getSessionUser?_t={networkTools.ts()}")

    try:
        res = res.json()
        assert res["code"] == 200
        res = res["data"]
        return models.Student(res["user"]["name"], res["uid"], res["user"]["sex"], res["user"]["faculty"],
                              res["user"]["facultyName"], res["user"]["grade"])
    except AssertionError or json.JSONDecodeError or KeyError:
        return None


def sessionLogout(sessionId=None, cookie=None, **kw) -> bool:
    """
    注销sessionID的登录状态。如果您正在使用SessionID类，则您并不需要调用本函数。
    @params: 在sessionId与cookie中任选一项传入即可。
    @return: 返回是否成功
    """

    if "session" not in kw:
        sessionId = _processArgs(sessionId, cookie)
        cookie = sessionId[1]
        sessionId = sessionId[0]

        uid = sessionIdToUserData(cookie=cookie).studentId
        res = requests.post("https://1.tongji.edu.cn/api/sessionservice/session/logout", json={
            "sessionid": sessionId,
            "uid": uid
        }, headers=networkTools.headers(), cookies=cookie)
    else:
        session: requests.Session = kw['session']
        uid: str = kw['uid']
        res = session.post("https://1.tongji.edu.cn/api/sessionservice/session/logout", json={
            "sessionid": sessionId,
            "uid": uid
        })

    try:
        res = res.json()
        assert res["code"] == 200
        return True
    except AssertionError or json.JSONDecodeError or KeyError:
        return False


"""
PART.2  各种信息类
"""


def getSchoolCalender(sessionId=None, cookie=None, **kw) -> Optional[dict]:
    """
    获取当前学期的校历。返回数据格式参考文档
    @params: 在sessionId与cookie中任选一项传入即可。
    @return: 字典格式的查询结果。若失败，则返回None。
    """
    if "session" not in kw:
        cookie = _processArgs(sessionId, cookie)[1]
        res = requests.get(
            f"https://1.tongji.edu.cn/api/baseresservice/schoolCalendar/currentTermCalendar?_t={networkTools.ts()}",
            cookies=cookie, headers=networkTools.headers())
    else:
        session: requests.Session = kw["session"]
        res = session.get(
            f"https://1.tongji.edu.cn/api/baseresservice/schoolCalendar/currentTermCalendar?_t={networkTools.ts()}")

    try:
        res = res.json()
        assert res["code"] == 200
        return res["data"]
    except AssertionError or json.JSONDecodeError or KeyError:
        return None


def getHolidayByYear(sessionId=None, cookie=None, year=time.localtime(time.time()).tm_year, **kw) -> Optional[dict]:
    """
    获取指定年份的假期安排。返回数据格式参考文档
    @params: 在sessionId与cookie中任选一项传入即可。year不填则为本年。
    @return: 字典格式的查询结果。若失败，则返回None。
    """
    if "session" not in kw:
        cookie = _processArgs(sessionId, cookie)[1]
        res = requests.get(
            f"https://1.tongji.edu.cn/api/baseresservice/holiday/queryHolidayByYear?year={year}&_t={networkTools.ts()}",
            cookies=cookie, headers=networkTools.headers())
    else:
        session: requests.Session = kw['session']
        res = session.get(
            f"https://1.tongji.edu.cn/api/baseresservice/holiday/queryHolidayByYear?year={year}&_t={networkTools.ts()}")

    try:
        res = res.json()
        assert res["code"] == 200
        return res["data"]
    except AssertionError or json.JSONDecodeError or KeyError:
        return None


def getScore(sessionId=None, cookie=None, **kw) -> Optional[models.Scores]:
    """
    获取本人成绩。返回Scores对象，原始字典在其data属性内。
    @params: 在sessionId与cookie中任选一项传入即可。
    @return: Scores对象。若失败，则返回None。
    """
    if "session" not in kw:
        cookie = _processArgs(sessionId, cookie)[1]
        uid = sessionIdToUserData(cookie=cookie).studentId
        res = requests.get(
            f"https://1.tongji.edu.cn/api/scoremanagementservice/scoreGrades/getMyGrades?studentId={uid}&_t={networkTools.ts()}",
            headers=networkTools.headers(), cookies=cookie)
    else:
        session: requests.Session = kw["session"]
        uid = kw["uid"]
        res = session.get(
            f"https://1.tongji.edu.cn/api/scoremanagementservice/scoreGrades/getMyGrades?studentId={uid}&_t={networkTools.ts()}")

    try:
        res = res.json()
        assert res["code"] == 200
        res = res["data"]
        return models.Scores(res)
    except AssertionError or json.JSONDecodeError or KeyError:
        return None


if __name__ == "__main__":
    print(repr(getScore(sessionId="821e209ee07848ef8995c7912679cfb4")))



"""
  Project Tongji-EasyAPI
  verifyTools.py
  Copyright (c) 2022 Cinea Zhan. All rights reserved
  www.cinea.com.cn
"""

import base64
import json

try:
    from . import networkTools, offlineCaptcha
except ImportError or ModuleNotFoundError:
    import networkTools
    import offlineCaptcha
import math
from random import random
import requests
from bs4 import BeautifulSoup
from Crypto.Cipher import AES
import Crypto.Util.Padding as Padding


# noinspection DuplicatedCode,SpellCheckingInspection
def genUUID():
    """
    本函数可以生成一个在登录验证中使用的UUID。如果你需要手动写通过验证码的过程，那么你可能会需要它。
    @param: 无需任何参数
    @return: 生成的UUID
    """
    s = []
    hexDigits = "0123456789abcdef"
    for i in range(36):
        s.append(hexDigits[math.floor(random() * 0x10)])
    s[14] = '4'
    s[19] = hexDigits[(int(s[19], base=16) & 0x3) | 0x8]
    s[8] = s[13] = s[18] = s[23] = "-"
    return "".join(s)


# noinspection SpellCheckingInspection
def updateSM2PublicKey(session: requests.Session = None):
    """
    本函数用于从ids.tongji.edu.cn获取最新的SM2公钥及其对应的cookie。如果你需要手动写通过验证码的过程，那么你可能会需要它。
    @param: 无需任何参数
    @return: 字典。secretId为公钥，cookie为对应的cookie（request.RequestCookieJar格式）。请注意，每个公钥都对应唯一的JSESSIONID。
    """
    if session:
        res = session.get(
            "https://ids.tongji.edu.cn:8443/nidp/app/login?id=Login&sid=0&option=credential&sid=0&target=https%3A%2F"
            "%2Fids.tongji.edu.cn%3A8443%2Fnidp%2Foauth%2Fnam%2Fauthz%3Fscope%3Dprofile%26response_type%3Dcode"
            "%26redirect_uri%3Dhttps%3A%2F%2F1.tongji.edu.cn%2Fapi%2Fssoservice%2Fsystem%2FloginIn%26client_id"
            "%3D5fcfb123-b94d-4f76-89b8-475f33efa194",
            headers=networkTools.idsHeaders())
        soup = BeautifulSoup(res.text, "lxml")
        secretId = soup.find(id="secretId")["value"]
    else:
        res = requests.get(
            "https://ids.tongji.edu.cn:8443/nidp/app/login?id=Login&sid=0&option=credential&sid=0&target=https%3A%2F"
            "%2Fids.tongji.edu.cn%3A8443%2Fnidp%2Foauth%2Fnam%2Fauthz%3Fscope%3Dprofile%26response_type%3Dcode"
            "%26redirect_uri%3Dhttps%3A%2F%2F1.tongji.edu.cn%2Fapi%2Fssoservice%2Fsystem%2FloginIn%26client_id"
            "%3D5fcfb123-b94d-4f76-89b8-475f33efa194",
            headers=networkTools.idsHeaders())
        soup = BeautifulSoup(res.text, "lxml")
        secretId = soup.find(id="secretId")["value"]
    return {"secretId": secretId, "cookie": res.cookies}


# noinspection PyBroadException
def getSM2js():
    try:
        res = requests.get("https://www.cinea.com.cn/static/sm2.js")
        if res.status_code == 200:
            return res.text
    except:
        pass
    try:
        res = requests.get("https://blogsources-1305284863.file.myqcloud.com/static/sm2.js")
        if res.status_code == 200:
            return res.text
    except:
        raise SystemError("无法获取运行必须的文件，请检查你的互联网连接")


BaseUrl = "https://ids.tongji.edu.cn:8443/nidp/app/login?sid=0&sid=0"


def aesEncrypt(data: str, key: str) -> str:
    """
    内部函数。调用时，key的有效性自理。
    """
    key = key.encode("utf-8")
    data = Padding.pad(data.encode("utf-8"), 16)
    aes = AES.new(key, AES.MODE_ECB)
    result = aes.encrypt(data)
    result = base64.b64encode(result)
    result = result.decode("utf-8")
    return result


def captchaBreaker(session: requests.Session, online: bool = False) -> tuple[bool, str]:
    """
    开发者说明：

    限于个人水平，目前无法完成自动打码
    """

    session.headers["content-type"] = "application/json;charset=UTF-8"
    pointList = []
    captchaMeta = {}

    # 阶段一：获取图片验证码
    def getCaptchaOnce(_uuid: str):
        result = session.post(BaseUrl + "/getCaptcha=1",
                              json={"captchaType": "clickWord", "clientUid": _uuid, "ts": networkTools.ts()}).text
        try:
            result = json.loads(result)
        except json.JSONDecodeError:
            raise SystemError("1系统拒绝了我们的请求。")
        if result["repCode"] == "0000":
            image = result["repData"]["originalImageBase64"]
            secretKey = result["repData"]["secretKey"]
            token = result["repData"]["token"]
            wordList = result["repData"]["wordList"]
            return {"image": image, "secretKey": secretKey, "token": token, "wordList": wordList}
        else:
            raise SystemError("1系统拒绝了我们的请求：" + result["repMsg"])

    # 阶段三：验证图片验证码
    def verifyCaptcha(pointData: list, _captchaMeta: dict, _uuid: str) -> bool:
        result = session.post(BaseUrl + "/checkCaptcha=1", json={
            "captchaType": "clickWord",
            "clientUid": _uuid,
            "pointJson": aesEncrypt(json.dumps(pointData).replace(' ', ''), _captchaMeta["secretKey"]) if _captchaMeta[
                "secretKey"] else json.dumps(pointData),
            "token": _captchaMeta["token"],
            "ts": networkTools.ts()
        })
        try:
            vRes = json.loads(result.text)
        except json.JSONDecodeError:
            raise SystemError("1系统拒绝了我们的请求。")
        if vRes["repCode"] == "0000":
            return True  # 已通过验证，无需处理返回内容
        else:
            return False

    # 阶段二：提交打码平台打码
    # 为便于日后维护，将由作者的服务器代理代码请求
    if online:
        # 在线打码
        success = False
        for _ in range(30):
            uuid = "point-" + genUUID()
            captchaMeta = getCaptchaOnce(uuid)
            try:
                res = requests.get("https://www.cinea.com.cn/api/sqp/captchaBroke", json={
                    "imageBase64": captchaMeta["image"],
                    "wordList": captchaMeta["wordList"]
                })
                if res.status_code != 200:
                    pass
                else:
                    resData = json.loads(res.text)
                    if resData["success"]:
                        pointList = []
                        for word in captchaMeta["wordList"]:
                            pointList.append({"x": resData["dataList"][word][0], "y": resData["dataList"][word][1]})
                        if verifyCaptcha(pointList, captchaMeta, uuid):
                            success = True
                            break
                        else:
                            requests.get(
                                f'https://www.cinea.com.cn/api/sqp/reportWrongCaptcha?picId={resData["picId"]}')
            except Exception as e:
                print("单次在线打码失败，原因：" + str(e))
                pass
        if not success:
            raise SystemError("很抱歉，验证码识别失败！在失败前已重试30次。")
    else:
        # 本地JavaScript打码（推荐）
        while True:
            uuid = "point-" + genUUID()
            captchaMeta = getCaptchaOnce(uuid)
            offlineResult = offlineCaptcha.captchaOfflineBreaker(captchaMeta["image"], captchaMeta["wordList"])
            if offlineResult["reason"] == 1:
                return False, ""
            pointList = offlineResult["result"]
            if verifyCaptcha(pointList, captchaMeta, uuid):
                break

    # 阶段四：制作验证码密钥
    captchaVerification = aesEncrypt(captchaMeta["token"] + '---' + json.dumps(pointList).replace(' ', ''),
                                     captchaMeta["secretKey"]) if captchaMeta["secretKey"] else captchaMeta["token"] + '---' + json.dumps(pointList).replace(' ', '')

    return True, captchaVerification



"""
  Project Tongji-EasyAPI
  models.pyi
  Copyright (c) 2022 Cinea Zhan. All rights reserved
  www.cinea.com.cn
"""

class Student:
    def __init__(self:Student,name:str|None="",studentId:str|None="",sex:int|None=0,faculty:str|None="",facultyName:str|None="",grade:str|None="",studentDataObject:dict|None=None) -> None:
        self.name:str
        self.studentId:str
        self.sex:int
        self.faculty:str
        self.facultyName:str
        self.grade:str
        ...
    
    def __str__(self:Student) -> str:...

    def __repr__(self:Student) -> str:...

class Scores:
    def __init__(self:Scores, sourceData:dict):
        self.gradePoint:float  #学生总寄点
        self.earnedCredits:float  #已修读学分
        self.failedCredits:float  #不及格学分
        self.failedNum:int  #不及格门数
        self.termNum:int   #学期数量
        self.coursesList:list     #课程列表

        self.courseNum:int  #已修读课程数量
        self.gradeExcellence:int   #获优课程门数
        self.gradeGood:int   #获良课程门数
        self.gradeMedium:int   #获中课程门数
        self.gradePass:int   #获及格课程门数
        self.gradeFailed:int   #获不及格课程门数
        ...

    def __str__(self:Scores) -> str:...

    def __repr__(self:Scores) -> str:...


"""
  Project Tongji-EasyAPI
  verifyTools.pyi
  Copyright (c) 2022 Cinea Zhan. All rights reserved
  www.cinea.com.cn
"""

import requests

def genUUID() -> str: ...

def updateSM2PublicKey(session:requests.Session) -> dict: ...

def getSM2js()->str: ...

def captchaBreaker(session: requests.Session, online: bool = False)-> tuple[bool,str]:...

def aesEncrypt(data: str, key: str) -> str:...




"""
  Project Tongji-EasyAPI
  session.pyi
  Copyright (c) 2022 Cinea Zhan. All rights reserved
  www.cinea.com.cn
"""

import threading
import requests

try:
    from . import models
except ImportError and ModuleNotFoundError:
    import models

class Session:  
    def __init__(self:Session,studentID:str|int|None=...,studentPassword:str|int|None=...,proxy:str|None=...)->bool:
        self.keepAlive:Session.keepAliveThread = None
        self.studentID:str = None
        self.islogin:bool = None
        self.id:int = None
        self.studentData:models.Student = None
        self.session:requests.Session = None
        self.keepaliveDestroy:threading.Event = None
        ...
    
    def testConnection(self:Session,url:str|None=...)->bool: ...

    def login(self:Session,studentID:str|int|None=...,studentPassword:str|int|None=...,cookie:str|None=...,manual:bool|None=...)->str: ...

    def request(self:Session,method:str,url:str,data:bytes|str|None=...,json:dict|list|None=...,params:dict|None=...)->requests.Response:...

    class keepAliveThread(threading.Thread):
        def __init__(self:Session.keepAliveThread, fatherName: str, waitTime: int, session:requests.Session, needDestroy: threading.Event)->None:
            self.session:requests.Session = None
            self.waitTime:int = None
            self.needDestroy:threading.Event = None
            self.id:int = None
            ...
        def run(self:Session.keepAliveThread)->None: ...


"""
  易用的一系统API。requests？F12找cookie？抱歉，忘了他们吧！
  ------------------------------------------------
  使用范例/快速上手：

  >>>  s = easyTongjiapi.Session()
  >>>  s.login(2152xxx,831xxx)
  >>>  print(s.sessionID)
  >>>  s.getScore()
  >>>  s.getScore().gradePoint
  -------------------------------------------------
  Project Tongji-EasyAPI
  __init__.py
  Copyright (c) 2022 Cinea Zhan. All rights reserved
  www.cinea.com.cn
"""

from .__version__ import __65472__ #大误
from .__version__ import __author__,__author_email__,__copyright__
from .__version__ import __title__,__description__,__url__
from .__version__ import __build__,__version__

from .session import Session
from .models import Student,Scores
from .function import *


"""
  Hey,don't import me! I am for internal use.

  主要是用到的几个结构，包括学生信息、课程信息、成绩信息等等
  -------------------------------------
  Project Tongji-EasyAPI
  models.py
  Copyright (c) 2022 Cinea Zhan. All rights reserved
  www.cinea.com.cn
"""


class Student:
    """
    学生信息类型，推荐用户使用
    """

    def __init__(self, name="", studentId="", sex=0, faculty="", facultyName="", grade="",
                 studentDataObject=None) -> None:
        self.sourceObj = None
        if studentDataObject:
            self.sourceObj = studentDataObject
            try:
                name = studentDataObject["name"]
                studentId = studentDataObject["uid"]
                sex = studentDataObject["sex"]
                faculty = studentDataObject["faculty"]
                facultyName = studentDataObject["facultyName"]
                grade = studentDataObject["grade"]
            except KeyError:
                raise ValueError("您传入的数据不正确，若您没有来自1系统的user对象，则请您不要传入此参数。")
        self.name = name
        self.studentId = studentId
        self.sex = sex
        self.faculty = faculty
        self.facultyName = facultyName
        self.grade = grade

    def __str__(self) -> str:
        return f"<{self.name},{self.studentId},{self.facultyName},{self.grade}级>"

    def __repr__(self) -> str:
        return str(self)


class Scores:
    """
    集中存储查询到的成绩的类。既可以存储原始字典，又能自动计算丰富的元数据。

    我们相信，这些元数据可以帮助你省去不必要的代码。

    请查阅文档获取详细的属性信息。
    """

    def __init__(self, sourceData: dict):
        """
        集中存储查询到的成绩的类。既可以存储原始字典，又能自动计算丰富的元数据。

        我们相信，这些元数据可以帮助你省去不必要的代码。

        请查阅文档获取详细的属性信息。

        请注意：你需要传入的是`data`字段的内容
        """
        sd = sourceData  # 节省代码量
        self.data = sourceData

        try:
            self.gradePoint = sd["totalGradePoint"]  # 学生总寄点
            self.earnedCredits = sd["actualCredit"]  # 已修读学分
            self.failedCredits = sd["failingCredits"]  # 不及格学分
            self.failedNum = sd["failingCourseCount"]  # 不及格门数
            self.termNum = len(sd["term"])  # 学期数量
            self.coursesList = []  # 课程列表

            courseNum = 0
            gradesNum = [0, 0, 0, 0, 0, 0]
            for term in sd["term"]:
                courseNum += len(term["creditInfo"])
                for courses in term["creditInfo"]:
                    gradesNum[int(courses["gradePoint"])] += 1
                    self.coursesList.append(courses)

            self.courseNum = courseNum  # 已修读课程数量
            self.gradeExcellence = gradesNum[5]  # 获优课程门数
            self.gradeGood = gradesNum[4]  # 获良课程门数
            self.gradeMedium = gradesNum[3]  # 获中课程门数
            self.gradePass = gradesNum[2]  # 获及格课程门数
            self.gradeFailed = gradesNum[0]  # 获不及格课程门数

            # 更多功能，发issue，都可以加！

        except KeyError as e:
            raise ValueError("传入的内容不能被正确解析，请检查是否完整、正确传入, 错误信息：" + str(e))

    def __str__(self):
        return str(self.gradePoint)

    def __repr__(self) -> str:
        return f"<gpa:{self.gradePoint}, terms:{self.termNum}, courses:{self.courseNum}, excellence:{self.gradeExcellence}>"



"""
  Project Tongji-EasyAPI
  function.pyi
  Copyright (c) 2022 Cinea Zhan. All rights reserved
  www.cinea.com.cn
"""

from . import models
from typing import Optional

def sessionIdToUserData(sessionId:str|None = ...,cookie:str|dict|None = ..., **kw) -> Optional[models.Student]: ...

def sessionLogout(sessionId:str|None = ...,cookie:str|dict|None = ..., **kw)->bool:...

def getSchoolCalender(sessionId:str|None = ...,cookie:str|dict|None = ..., **kw) -> Optional[dict]:...

def getHolidayByYear(sessionId:str|None = ...,cookie:str|dict|None = ...,year:int|None=...,**kw) -> Optional[dict]:...

def getScore(sessionId:str|None = ...,cookie:str|dict|None = ..., **kw)->Optional[models.Scores]:...






from setuptools import setup, find_packages
import pathlib, os

here = pathlib.Path(__file__).parent.resolve()
long_description = (here / "README.md").read_text(encoding="utf-8")

if os.name == "posix" and os.system("gcc --version") != 0:
    print("警告：没有在您的电脑上发现gcc，安装可能不成功")

setup(
    name="easyTongjiapi",
    version="0.1.0a5",
    description="Easy Tongji 1-System API for everyone.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Cinea4678/Tongji-EasyAPI",
    author="Cinea Zhan",
    author_email="cinea@cinea.com.cn",
    license="GPL v3.0",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 2 - Pre-Alpha",
        "Natural Language :: Chinese (Simplified)"
    ],
    package_dir={"": "src"},
    packages=find_packages(where="src"),
    python_requires=">=3",
    install_requires=["requests", "beautifulsoup4", "lxml", "fastgm-whl",
                      "opencv-python", "pycryptodome", "pillow", "numpy"],
    project_urls={
        "Bug Reports": "https://github.com/Cinea4678/Tongji-EasyAPI/issues",
        "Visit My Homepage": "https://www.cinea.com.cn",
        "Source": "https://github.com/Cinea4678/Tongji-EasyAPI",
    },
)



"""
  Project Tongji-EasyAPI
  crack.pyi
  Origin Project : https://github.com/seashell11234455/crackids  License GPL v3.0
  Origin Author:  FeO3  https://github.com/seashell11234455
  www.cinea.com.cn
"""

def getCode()->str: ...



"""
  Keep love for life! 
  ~~~~~~~~~~~~~~~~~~~~~~~~~~

  Project Tongji-EasyAPI
  __version__.py
  Copyright (c) 2022 Cinea Zhan. All rights reserved
  www.cinea.com.cn
"""

__title__ = 'easyTongjiapi'
__description__ = 'Easy Tongji 1-System API for everyone.'
__url__ = 'https://www.cinea.com.cn'
__version__ = '0.1.0'
__build__ = 0x001000
__author__ = 'Cinea Zhan'
__author_email__ = 'cinea@cinea.com.cn'
__license__ = 'GPL v3.0'
__copyright__ = 'Copyright 2022 Cinea Zhan'
__65472__ = 'RRRRRR\nBBBBB\nRRRR\nBBBBBBB\nRR\n分流政策是SB'


"""
  Project Tongji-EasyAPI
  session.py
  Copyright (c) 2022 Cinea Zhan. All rights reserved
  www.cinea.com.cn
"""

try:
    from . import verifyTools, networkTools, crack, models, function
except ImportError or ModuleNotFoundError:
    import verifyTools, networkTools, crack, models, function
import json
import random
import re
import threading
import time
import urllib.parse as urlparse

import requests


class Session:
    """
    同济大学教务系统的连接会话。
    """

    def __init__(self, studentID=None, studentPassword=None, manual=False, proxy=None):
        """
        初始化连接会话，*您可以选择立即登录至一系统，也可以在创建会话后登录。*
        @param studentID: 若现在登录，则为登录学生的**学号**。
        @param studentPassword: 若现在登录，则为登录学生的**密码**（推荐密码使用str）。
        @param manual: 是否手动输入验证码。若自动输入不成功，请尝试使用手动输入。
        @param proxy: 若需使用HTTP代理，请在此填入代理地址。
        @return: 构造函数无返回值
        """
        self.id = random.randint(1, 9999)
        self.islogin = False
        self.studentID = None
        self.studentPassword = None
        self.token = None
        self.uid = None
        self.sessionID = None
        self.session = requests.session()
        self.studentData = models.Student(name="对象创建成功")

        self.loginTime = 0

        self.keepaliveDestroy = threading.Event()  # 需要keep-alive进程销毁
        self.keepaliveDestroy.clear()
        self.keepAlive = self.keepAliveThread(str(self.id), 30, self.session, self.keepaliveDestroy)

        if proxy:
            if not re.match(r"^(https?|socks5?)://([^:]*(:[^@]*)?@)?([^:]+|\[[:0-9a-fA-F]+])(:\d+)?/?$|^$", proxy):
                raise ValueError(
                    "代理地址格式错误，必须匹配^(https?|socks5?)://([^:]*(:[^@]*)?@)?([^:]+|\\[[:0-9a-fA-F]+\\])(:\\d+)?/?$|^$")
            proxy = re.findall(r"(https?|socks5?)://(.*)", proxy)
            try:
                self.session.proxies = {proxy[0]: proxy[1]}
            except Exception as e:
                raise ValueError("代理地址不能被程序解析。异常原文：" + str(e))

        # 登录
        if studentID or studentPassword:
            self.login(studentID, studentPassword, manual=manual)

    def __del__(self):
        if self.islogin:
            self.logout()

    def testConnection(self, url=None):
        """
        测试连接会话是否已经成功建立。
        @params url: 你可以自定义测试连接时所前往连接的Url（请求方式为get，敬请注意）
        @return: 连接成功与否
        """
        if not self.islogin:
            return False  # 此处为假会导致绝大多数功能异常，不能允许用户测试
        if not self.studentID:
            return False
        if not url:
            url = f"https://1.tongji.edu.cn/api/studentservice/studentDetailInfo/getStatusInfoByStudentId?studentId={self.studentID}&_t={networkTools.ts()}"

        # 开始测试连接
        try:
            response = self.session.get(url)
            if response.status_code != 200:
                return False
            res = json.loads(response.text)
            if "code" not in res or res["code"] != 200:
                return False
            else:
                return True
        except json.JSONDecodeError:
            return False
        except Exception as e:
            raise e

    def login(self, studentID=None, studentPassword=None, cookie=None, manual=False):
        """
        登录至一系统。
        **学号密码登录**：传入学号与密码即可。studentID: 学号; studentPassword: 密码。
        @params manual:  参数manual用来确认使用自动通过验证码或手动通过验证码。如果您不能成功使用自动模式，请使用手动模式。手动模式仍然存在bug。
        **cookie登录**：传入cookie即可。
        @return: 函数会返回cookie，但是会话对象会自动登录，你无需额外操作。
        """

        # 参数完整性检查
        if cookie:
            # 优先使用cookie
            if not studentID:
                raise ValueError("未提供studentID")
        else:
            if not studentID or not studentPassword:
                raise ValueError("未完整提供studentID与studentPassword")

        # 使用cookie登录
        if cookie:
            if isinstance(studentID, int):
                studentID = str(studentID)
            # noinspection RegExpDuplicateCharacterInClass
            if not re.match(r"^[1|2|3|4][0-9]{6}$", studentID):
                raise ValueError("学号格式错误，必须匹配^[1|2|3|4][0-9]{6}$。")
            if isinstance(cookie, str):
                cookie: dict = networkTools.parseStrCookie(cookie)
            self.session.cookies = cookie
            self.session.headers = networkTools.headers()
            self.studentData = function.sessionIdToUserData(cookies=cookie)
            if self.studentData:
                self.sessionID = cookie["sessionid"]
                self.islogin = True
                self.studentID = self.studentData.studentId
                self.loginTime = time.time()
                self.keepAlive.setName(str(self.studentID))
                self.keepaliveDestroy.clear()
                self.keepAlive.start()
                return cookie
            else:
                raise ValueError("您提供的cookie不能正确使用！")

        # 学号与密码效验
        if isinstance(studentID, int):
            studentID = str(studentID)
        if isinstance(studentPassword, int):
            studentPassword = "%06d" % studentPassword  # 补齐零
        if not re.match("^[1|2|3|4][0-9]{6}$", studentID):
            raise ValueError("学号格式错误，必须匹配^[1|2|3|4][0-9]{6}$。")
        if not re.match("^[0-9]{6}$", studentPassword):
            raise ValueError("密码格式错误，必须匹配^[0-9]{6}$。")

        # 初始化Session
        self.session.headers = networkTools.idsHeaders()

        # 加密密码
        IDSSM2PublicKey = verifyTools.updateSM2PublicKey(self.session)["secretId"]
        encryptData = networkTools.sm2Encrypt(studentPassword, IDSSM2PublicKey)

        # 完成验证码验证
        captchaVerification = ""
        if manual:
            # 手动完成验证
            captchaRes = verifyTools.captchaBreaker(self.session, False)
            if not captchaRes[0]:
                raise KeyboardInterrupt("用户手动结束了操作！")
            captchaVerification = captchaRes[1]

        for _ in range(1 if manual else 1):

            if not manual:
                # 自动验证
                captchaVerification = crack.getCode()

            # 提交验证，开始登录
            # 第一跳：向ids取得访问权限
            # noinspection SpellCheckingInspection
            dataToSend = {
                "option": "credential",
                "Ecom_Captche": captchaVerification,
                "Ecom_User_ID": studentID,
                "Ecom_Password": encryptData
            }
            dataToSend = urlparse.urlencode(dataToSend)
            self.session.headers["content-type"] = "application/x-www-form-urlencoded"
            loginResp1 = self.session.post("https://ids.tongji.edu.cn:8443/nidp/app/login?sid=0&sid=0", data=dataToSend)
            urls = re.findall(r"window\.location\.href=\'(.*?)\'", loginResp1.text)
            if len(urls) > 0:
                href = urls[0]
            else:
                raise ValueError("登录失败，请检查学号，密码是否正确！")

            # 第二三四跳：由ids前往1系统
            loginResp2 = self.session.get(href, headers=networkTools.idsHeaders())

            # 第五跳：向1系统取得cookies
            try:
                self.session.headers = networkTools.headers()
                self.session.headers["x-token"] = ""
                tokenForm = urlparse.parse_qs(urlparse.urlparse(loginResp2.url).query)
                token = tokenForm["token"][0]
                uid = tokenForm["uid"][0]
                FuckTheFakeTS = tokenForm["ts"][0]
                loginResp3 = self.session.post(f"https://1.tongji.edu.cn/api/sessionservice/session/login",
                                               data=json.dumps({
                                                   "uid": uid,
                                                   "token": token,
                                                   "ts": FuckTheFakeTS
                                               }).replace(' ', ''))
            except Exception as e:
                if not manual:
                    continue
                else:
                    raise e  # 手动登录是不会出现这种问题滴！

            try:
                loginResult = json.loads(loginResp3.text)
                if "code" not in loginResult or loginResult["code"] != 200:
                    raise SystemError(
                        f"登录失败，1系统返回了不正常的凭据。这通常是由于访问人数过多造成的。频繁出现此错误，则请联系开发者。（notes: json loads succeed but not code 200. The text is '{loginResp3.text}'）")
                loginResult = loginResult["data"]
                self.sessionID = loginResult["sessionid"]
                studentDataSourceObj = loginResult["user"]
                self.studentData = models.Student(studentDataObject=studentDataSourceObj)
            except:
                raise SystemError(
                    f"登录失败，1系统返回了不正常的凭据。这通常是由于访问人数过多造成的。频繁出现此错误，则请联系开发者。（notes: json loads failed while text is '{loginResp3.text}'）")

            # 测试连接
            self.islogin = True
            self.studentID = studentID
            self.studentPassword = studentPassword
            if self.testConnection():
                self.loginTime = time.time()
                self.keepAlive.setName(str(self.studentID))
                self.keepaliveDestroy.clear()
                self.keepAlive.start()
                return self.session.cookies.get_dict()
            else:
                self.islogin = False
                raise SystemError("登录失败，1系统未正常运行")

        if not manual:
            raise SystemError("自动验证失效，这真是不可思议。请使用手动模式通过验证码。")

    class keepAliveThread(threading.Thread):
        """
        同济大学教务系统Session的Keep-alive线程，请不要自行使用。
        """

        def __init__(self, fatherName: str, waitTime: int, session: requests.Session, needDestroy: threading.Event):
            self.waitTime = waitTime
            self.session = session
            self.needDestroy = needDestroy
            self.id = random.randint(1, 9999)
            super().__init__(name=f"1-Session-{fatherName} KeepAliveThread {self.id}")

        def setName(self, name: str) -> None:
            return super().setName(f"1-Session-{name} KeepAliveThread {self.id}")

        def run(self):
            lastRun = 0
            while not self.needDestroy.is_set():
                if time.time() - lastRun > self.waitTime:
                    self.session.get(
                        f"https://1.tongji.edu.cn/api/baseresservice/schoolCalendar/currentTermCalendar?_t={networkTools.ts()}")
                    lastRun = time.time()
                time.sleep(2)

    def request(self, method, url, data=None, _json=None, params=None) -> requests.Response:
        if not self.islogin:
            raise SystemError("请先登录再使用此功能")
        return self.session.request(method=method, url=url, params=params, data=data, json=_json)

    def logout(self) -> bool:
        if not self.islogin:
            raise SystemError("请先登录再使用此功能")
        self.islogin = not function.sessionLogout(sessionId=self.sessionID, session=self.session, uid=self.studentID)
        if not self.islogin:
            self.keepaliveDestroy.set()
        return not self.islogin

    # 以下是APIs
    def getSchoolCalender(self) -> dict:
        """
        获取当前学期的校历。返回数据格式参考文档
        @return: 字典格式的查询结果。若失败，则返回None。
        """
        if not self.islogin:
            raise SystemError("请先登录再使用此功能")
        return function.getSchoolCalender(session=self.session)

    def getHolidayByYear(self, year=time.localtime(time.time()).tm_year) -> dict:
        """
        获取指定年份的假期安排。返回数据格式参考文档
        @params: 年份。不填则为本年
        @return: 字典格式的查询结果。若失败，则返回None。
        """
        if not self.islogin:
            raise SystemError("请先登录再使用此功能")
        return function.getHolidayByYear(year=year, session=self.session)

    def getScore(self) -> models.Scores:
        """
        获取本人成绩。返回Scores对象，原始字典在其data属性内。
        @return: Scores对象。若失败，则返回None。
        """
        if not self.islogin:
            raise SystemError("请先登录再使用此功能")
        return function.getScore(uid=self.studentID, session=self.session)

    def __str__(self) -> str:
        return self.studentData.name

    def __repr__(self) -> str:
        return f"<TJU 1-Session, User:{str(self.studentData)}, Last Login Time:{time.asctime(time.localtime(self.loginTime))}, sessionID:{self.sessionID}>"


if __name__ == "__main__":
    print(Session("2152955", "831033"))



"""
  Hey,don't import me! I am for internal use.
  -------------------------------------
  Project Tongji-EasyAPI
  networkTools.py
  Copyright (c) 2022 Cinea Zhan. All rights reserved
  www.cinea.com.cn
"""

import time, math
from random import random
import base64
import fastgm


def _genUUID():
    """
    本函数可以生成一个在登录验证中使用的UUID。如果你需要手动写通过验证码的过程，那么你可能会需要它。
    @param: 无需任何参数
    @return: 生成的UUID
    """
    s = []
    hexDigits = "0123456789abcdef"
    for i in range(36):
        s.append(hexDigits[math.floor(random() * 0x10)])
    s[14] = '4'
    s[19] = hexDigits[(int(s[19], base=16) & 0x3) | 0x8]
    s[8] = s[13] = s[18] = s[23] = "-"
    return "".join(s)


# noinspection SpellCheckingInspection
_idsHeaders = {
    "accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9",
    "accept-encoding": "gzip, deflate, br",
    "accept-language": "zh-CN,zh;q=0.9",
    "Authorization": "Basic MjE1Mjk1NTo4MzEwMzM=",
    "origin": "https://ids.tongji.edu.cn:8443",
    "referer": "https://ids.tongji.edu.cn:8443/nidp/app/login?id=Login&sid=0&option=credential&sid=0&target=https%3A%2F%2Fids.tongji.edu.cn%3A8443%2Fnidp%2Foauth%2Fnam%2Fauthz%3Fscope%3Dprofile%26response_type%3Dcode%26redirect_uri%3Dhttps%3A%2F%2F1.tongji.edu.cn%2Fapi%2Fssoservice%2Fsystem%2FloginIn%26client_id%3D" + _genUUID(),
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.0.0 Safari/537.36"
}


def idsHeaders() -> dict:
    return _idsHeaders


# noinspection SpellCheckingInspection
_headers = {
    "accept": "application/json, text/plain, */*",
    "accept-encoding": "gzip, deflate, br",
    "accept-language": "zh-CN,zh;q=0.9",
    "content-type": "application/json",
    "origin": "https://1.tongji.edu.cn",
    "referer": "https://1.tongji.edu.cn/workbench",
    "user-agent": "Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/104.0.0.0 Safari/537.36"
}


def headers() -> dict:
    return _headers


def ts() -> int:
    """Time stamp of JavaScript"""
    return int(time.time() * 1000)


def sm2Encrypt(data: str, publicKey: str):
    """
    不要在ids登录验证以外的场景使用本函数！！！
    """
    sm2c = fastgm.SM2("C1C2C3")
    data_enc = base64.b64encode(data.encode("utf-8"))
    return "04" + sm2c.encrypt(publicKey[2:], data_enc).hex()


def parseStrCookie(cookie):
    return dict([l.split("=", 1) for l in cookie.split("; ")])


if __name__ == '__main__':
    # noinspection SpellCheckingInspection
    print(sm2Encrypt("CINEA!",
                     "3957923131D9E2F93797869D0F03EEA3268B6F490858057AA696A0BE0A4E94E71AD54761A7205A42D3C63F78DC3EAFE6F48088BAEE5D5654C70EE12F80B62F3F"))



"""
  This file is planned to be deployed on one cloud server so that users don't  have to install OpenCV, Pillows and so on.
  感谢刘孔阳大佬
  --------------------------------------------
  Project Tongji-EasyAPI
  crack.py
  Origin Project : https://github.com/seashell11234455/crackids  License GPL v3.0
  Origin Author:  FeO3  https://github.com/seashell11234455
  www.cinea.com.cn
"""

import os
import requests
import cv2
import base64
import numpy as np
from Crypto.Cipher import AES
import base64
import json
from PIL import Image, ImageDraw, ImageFont


def getData():
    url = "https://ids.tongji.edu.cn:8443/nidp/app/login?sid=0&sid=0/getCaptcha=1"
    resp = requests.post(url)
    return resp.json()["repData"]


def check(data, point):
    url = "https://ids.tongji.edu.cn:8443/nidp/app/login?sid=0&sid=0/checkCaptcha=1"
    enc = encrypt(json.dumps(point).replace(" ", ""), data["secretKey"])
    resp = requests.post(url,json={ "token": data["token"], "pointJson": enc })
    return resp.json()


def encrypt(s, key):
    def pad(m):
        left = 16 - len(m) % 16
        return m + chr(left) * left

    cipher = AES.new(key.encode("utf-8"), AES.MODE_ECB)
    enc_byte = cipher.encrypt(pad(s).encode("utf-8"))
    return base64.b64encode(enc_byte).decode("ascii")


def decrypt(s, key):
    cipher = AES.new(key.encode("utf-8"), AES.MODE_ECB)
    b = cipher.decrypt(base64.b64decode(s))
    return b[: -b[-1]]


def getImageFromBase64(b64):
    buffer = base64.b64decode(b64)
    nparr = np.frombuffer(buffer, np.uint8)
    image = cv2.imdecode(nparr, cv2.IMREAD_COLOR)
    return image


def findContour(img):
    if cv2.__version__.startswith('3'):
        _, contours, _ = cv2.findContours(img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)
    else: 
        contours, _ = cv2.findContours(img, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_SIMPLE)

    def find_if_close(cnt1, cnt2):
        row1, row2 = cnt1.shape[0], cnt2.shape[0]
        for i in range(row1):
            for j in range(row2):
                dist = np.linalg.norm(cnt1[i] - cnt2[j])
                if abs(dist) < 2:
                    return True
                elif i == row1 - 1 and j == row2 - 1:
                    return False

    LENGTH = len(contours)
    status = np.zeros((LENGTH, 1))
    for i, cnt1 in enumerate(contours):
        x = i
        if i != LENGTH - 1:
            for j, cnt2 in enumerate(contours[i + 1 :]):
                x = x + 1
                dist = find_if_close(cnt1, cnt2)
                if dist == True:
                    val = min(status[i], status[x])
                    status[x] = status[i] = val
                else:
                    if status[x] == status[i]:
                        status[x] = i + 1
    unified = []
    maximum = int(status.max()) + 1
    for i in range(maximum):
        pos = np.where(status == i)[0]
        if pos.size != 0:
            cont = np.vstack([contours[i] for i in pos])
            hull = cv2.convexHull(cont)
            unified.append(hull)
    return unified


def extractChar(img, contour):
    mult = 1.1
    ret = []
    pt = []
    for cnt in contour:
        rect = cv2.minAreaRect(cnt)
        box = cv2.boxPoints(rect)
        box = np.int0(box)

        W = rect[1][0]
        H = rect[1][1]
        if W * H < 10:
            continue

        Xs = [i[0] for i in box]
        Ys = [i[1] for i in box]
        x1 = min(Xs)
        x2 = max(Xs)
        y1 = min(Ys)
        y2 = max(Ys)

        rotated = False
        angle = rect[2]

        if angle < -45:
            angle += 90
            rotated = True

        center = (int((x1 + x2) / 2), int((y1 + y2) / 2))
        size = (int(mult * (x2 - x1)), int(mult * (y2 - y1)))

        M = cv2.getRotationMatrix2D((size[0] / 2, size[1] / 2), angle, 1.0)

        cropped = cv2.getRectSubPix(img, size, center)
        cropped = cv2.warpAffine(cropped, M, size)

        croppedW = W if not rotated else H
        croppedH = H if not rotated else W

        croppedRotated = cv2.getRectSubPix(
            cropped,
            (int(croppedW * mult), int(croppedH * mult)),
            (size[0] / 2, size[1] / 2),
        )

        if croppedRotated.size < 200:
            continue
        if np.mean(croppedRotated) < 30:
            continue
        im = cv2.resize(croppedRotated, (30, 30))
        kernel = np.array([[0, -1, 0], [-1, 5, -1], [0, -1, 0]], np.float32)
        im = cv2.filter2D(im, -1, kernel=kernel)
        ret.append(im)
        pt.append((rect[0][0], rect[0][1]))
    return ret, pt


def genCharacter(ch, size):
    img = Image.new("L", size, 0)
    font = ImageFont.truetype("simsun.ttc", min(size))
    draw = ImageDraw.Draw(img)
    draw.text((0, 0), ch, font=font, fill=255)
    return np.asarray(img)

def crack(data):
    img = getImageFromBase64(data["originalImageBase64"])
    img = cv2.cvtColor(img, cv2.COLOR_BGR2GRAY)
    _, img = cv2.threshold(img, 1, 255, cv2.THRESH_BINARY)
    img = cv2.bilateralFilter(img, 5, 200, 200)
    img = cv2.bitwise_not(img)
    cnt = findContour(img)
    ch, pt = extractChar(img, cnt)
    wd = [genCharacter(w, (30, 30)) for w in data["wordList"]]
    score = []
    for i, w in enumerate(wd):
        for j, c in enumerate(ch):
            score.append(((i, j), cv2.bitwise_xor(c, w).sum()))
    score.sort(key=lambda x: x[1])
    wset = set()
    cset = set()
    ans = {}

    for s in score:
        if s[0][0] in wset or s[0][1] in cset:
            continue
        wset.add(s[0][0])
        cset.add(s[0][1])
        ans[data["wordList"][s[0][0]]] = pt[s[0][1]]

    return [{"x": int(ans[w][0]), "y": int(ans[w][1])} for w in data["wordList"]]


def getCode():
    for i in range(20):
        data = getData()
        point = crack(data)
        if check(data, point)["repCode"] == "0000":
            raw = data["token"] + "---" + json.dumps(point).replace(" ", "")
            return encrypt(raw, data["secretKey"])
    return ""

if __name__ == "__main__":
    print(getCode(),end="")



"""
  Hey,don't import me! I am for internal use.
  -------------------------------------
  Project Tongji-EasyAPI
  offlineCaptcha.py
  Copyright (c) 2022 Cinea Zhan. All rights reserved
  www.cinea.com.cn
"""

#  不幸的是，使用打码平台处理ids系统的验证码并不是一个明智的选择
#  因此编写了这个offlineCaptcha系统，希望能通过用户手动点选以完成1系统的验证


from http.server import HTTPServer, BaseHTTPRequestHandler
import os, threading, json, re, time, webbrowser
from urllib.parse import unquote
import inspect, ctypes


# 进程停止代码

def _async_raise(tid, exctype):
    """raises the exception, performs cleanup if needed"""
    tid = ctypes.c_long(tid)
    if not inspect.isclass(exctype):
        exctype = type(exctype)
    res = ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, ctypes.py_object(exctype))
    if res == 0:
        raise ValueError("invalid thread id")
    elif res != 1:
        """
        if it returns a number greater than one, you're in trouble,
        and you should call it again with exc=NULL to revert the effect
        """
        ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, None)
        raise SystemError("PyThreadState_SetAsyncExc failed")


def stop_thread(thread):
    _async_raise(thread.ident, SystemExit)


host = ('127.0.0.1', 810)
httpfilepath = os.path.dirname(os.path.abspath(__file__)) + "/web"  # web文件路径

# 使用全局变量通信
stopFlag = False
heartBeat = False
stopReason = 0  # 0:自动中断；1:用户中断
imageB64 = ""
wordList = ""
result = []


# 本地Server

class localCaptchaServer(BaseHTTPRequestHandler):
    finished = False

    # GET处理文件请求
    def do_GET(self):
        path = str(self.path)
        try:
            with open(httpfilepath + path, "rb") as file:
                fileContent = file.read()
            self.send_response(200)
            self.send_header("content-type", "*/*")
            self.end_headers()
            self.wfile.write(fileContent)
        except FileNotFoundError:
            self.send_error(404, "File Not Found")

    # POST处理api请求
    def do_POST(self):
        global stopFlag, stopReason
        path = str(self.path)
        command = re.findall(r"^/api/(.*?)$", path)
        if len(command) == 0:
            self.send_error(400, "Bad Request")
        else:
            command = command[0]
            if command == "hello":
                # 开发调试用
                self.send_response(200)
                self.send_header("content-type", "text/plain")
                self.end_headers()
                self.wfile.write("Hello!".encode())
            elif command == "heartbeat":
                self.send_response(200)
                self.send_header("content-type", "text/plain")
                self.end_headers()
                self.wfile.write("Ok!".encode())
                if not self.finished:
                    stopFlag = False
            elif command == "getdata":
                self.send_response(200)
                self.send_header("content-type", "application/json")
                self.end_headers()
                self.wfile.write(json.dumps({"imageBase64": imageB64, "wordsList": wordList}).encode())
            elif "submit" in command:
                command = unquote(command)
                jsonData = re.findall(r"json=(.*)", command)
                try:
                    jsonData = json.loads(jsonData[0])
                    global result
                    result = jsonData["result"]
                    self.send_response(200)
                    self.send_header("content-type", "text/plain")
                    self.end_headers()
                    self.wfile.write("ok".encode())
                    self.finished = True
                    stopFlag = True
                    stopReason = 0
                except json.JSONDecodeError or IndexError or KeyError:
                    self.send_error(400, "Bad request")
            else:
                self.send_error(404, "Command not found")


class checkFinishThread(threading.Thread):
    def __init__(self, server):
        self.server = server
        super().__init__()

    def run(self):
        global stopFlag
        time.sleep(5)  # 等待浏览器和http server启动
        while not stopFlag:
            stopFlag = True
            time.sleep(1)
        self.server.shutdown()


class openBrowserThread(threading.Thread):
    """
    使用线程打开浏览器的原因：希望能够延时几秒防止打开浏览器后后端服务未启动导致404
    """

    def __init__(self):
        self.opened = False
        super().__init__()

    def run(self):
        if not self.opened:
            time.sleep(1.5)
            webbrowser.open("http://127.0.0.1:810/index.html", new=1)
            self.opened = True


# 初始化/重新初始化app
# noinspection PyStatementEffect
def captchaOfflineBreaker(imagebs64: str, clickwords: list) -> dict:
    """
    @return: 返回一个字典。字段"result":结果，"reason"退出原因，为1时需中断验证过程并切断登陆。
    """
    # 重置全局变量
    print("call")
    global stopFlag, imageB64, wordList, result
    stopFlag = False
    imageB64 = imagebs64
    wordList = ",".join(clickwords)
    result = []

    server = HTTPServer(host, localCaptchaServer)
    checking = checkFinishThread(server)
    browserThread = openBrowserThread()
    browserThread.start()
    checking.start()
    server.serve_forever()

    # 检查结果有效性
    try:
        result[0]["x"], result[0]["y"]
        result[1]["x"], result[1]["y"]
        result[2]["x"], result[2]["y"]
    except IndexError or KeyError:
        result = [{"x": 0, "y": 0}, {"x": 0, "y": 0}, {"x": 0, "y": 0}]

    # stop_thread(browserThread)
    # stop_thread(checking)

    return {"result": result, "reason": stopReason}


if __name__ == "__main__":
    with open("image.jpg", "rb") as f:
        imageB64 = f.read()
    import base64

    imageB64 = base64.b64encode(imageB64).decode("utf-8")
    captchaOfflineBreaker(imageB64, ["我", "尼", "马"])
